package dev.shadowsoffire.apotheosis.adventure.commands;

import java.util.Objects;
import java.util.Collections;
import java.util.stream.Stream;
import net.minecraft.commands.SharedSuggestionProvider;
import java.util.function.Function;
import java.util.concurrent.CompletableFuture;
import com.mojang.brigadier.suggestion.SuggestionsBuilder;
import com.mojang.brigadier.exceptions.CommandSyntaxException;
import java.util.List;
import dev.shadowsoffire.attributeslib.api.AttributeHelper;
import java.util.Set;
import dev.shadowsoffire.apotheosis.adventure.loot.LootController;
import net.minecraft.network.chat.MutableComponent;
import java.util.Map;
import net.minecraft.network.chat.HoverEvent;
import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.Style;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.entity.Entity;
import net.minecraft.network.chat.Component;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixInstance;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixHelper;
import net.minecraft.world.entity.LivingEntity;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixRegistry;
import net.minecraft.resources.ResourceLocation;
import com.mojang.brigadier.context.CommandContext;
import com.mojang.brigadier.builder.ArgumentBuilder;
import com.mojang.brigadier.arguments.FloatArgumentType;
import com.mojang.brigadier.arguments.ArgumentType;
import net.minecraft.commands.arguments.ResourceLocationArgument;
import com.mojang.brigadier.builder.RequiredArgumentBuilder;
import net.minecraft.commands.Commands;
import com.mojang.brigadier.builder.LiteralArgumentBuilder;
import net.minecraft.commands.CommandSourceStack;
import com.mojang.brigadier.suggestion.SuggestionProvider;

public class AffixCommand
{
    public static final SuggestionProvider<CommandSourceStack> SUGGEST_AFFIX;
    public static final SuggestionProvider<CommandSourceStack> SUGGEST_APPLICABLE_AFFIX;
    public static final SuggestionProvider<CommandSourceStack> SUGGEST_AFFIX_ON_ITEM;
    
    public static void register(final LiteralArgumentBuilder<CommandSourceStack> root) {
        final LiteralArgumentBuilder<CommandSourceStack> builder = (LiteralArgumentBuilder<CommandSourceStack>)Commands.m_82127_("affix").requires(c -> c.m_6761_(2));
        builder.then(Commands.m_82127_("apply").then(((RequiredArgumentBuilder)Commands.m_82129_("affix", (ArgumentType)ResourceLocationArgument.m_106984_()).suggests((SuggestionProvider)AffixCommand.SUGGEST_APPLICABLE_AFFIX).then(Commands.m_82129_("level", (ArgumentType)FloatArgumentType.floatArg(0.0f, 1.0f)).executes(c -> applyAffix((CommandContext<CommandSourceStack>)c, ResourceLocationArgument.m_107011_(c, "affix"), FloatArgumentType.getFloat(c, "level"))))).executes(c -> applyAffix((CommandContext<CommandSourceStack>)c, ResourceLocationArgument.m_107011_(c, "affix"), ((CommandSourceStack)c.getSource()).m_81372_().f_46441_.m_188501_()))));
        builder.then(Commands.m_82127_("list").executes(AffixCommand::listAffixes));
        builder.then(Commands.m_82127_("list_alternatives").then(Commands.m_82129_("affix", (ArgumentType)ResourceLocationArgument.m_106984_()).suggests((SuggestionProvider)AffixCommand.SUGGEST_AFFIX_ON_ITEM).executes(c -> listAlternatives((CommandContext<CommandSourceStack>)c, ResourceLocationArgument.m_107011_(c, "affix")))));
        root.then((ArgumentBuilder)builder);
    }
    
    public static int applyAffix(final CommandContext<CommandSourceStack> c, final ResourceLocation affixId, final float level) {
        final DynamicHolder<Affix> afx = (DynamicHolder<Affix>)AffixRegistry.INSTANCE.holder(affixId);
        if (!afx.isBound()) {
            return fail(c, "Unknown affix: " + affixId, -1);
        }
        final Entity entity = ((CommandSourceStack)c.getSource()).m_81373_();
        if (!(entity instanceof LivingEntity)) {
            return fail(c, "/apoth affix must be executed by a living entity.", -10);
        }
        final LivingEntity living = (LivingEntity)entity;
        final ItemStack held = living.m_21205_();
        if (held.m_41619_()) {
            return fail(c, "The target entity must have an item in their main hand.", -2);
        }
        final DynamicHolder<LootRarity> rarity = (DynamicHolder<LootRarity>)AffixHelper.getRarity(held);
        if (!rarity.isBound()) {
            return fail(c, "The target item must have a set rarity.", -3);
        }
        final LootCategory cat = LootCategory.forItem(held);
        if (cat.isNone()) {
            return fail(c, "The target item must have a valid loot category", -4);
        }
        if (!((Affix)afx.get()).canApplyTo(held, cat, (LootRarity)rarity.get())) {
            return fail(c, "The selected affix cannot be applied to the target item.", -5);
        }
        AffixHelper.applyAffix(held, new AffixInstance((DynamicHolder)afx, held, (DynamicHolder)rarity, level));
        ((CommandSourceStack)c.getSource()).m_288197_(() -> Component.m_237110_("Successfully applied affix %s with level %s to %s", new Object[] { affixId, level, held.m_41611_() }), true);
        return 0;
    }
    
    public static int listAffixes(final CommandContext<CommandSourceStack> c) {
        final Entity entity = ((CommandSourceStack)c.getSource()).m_81373_();
        if (!(entity instanceof LivingEntity)) {
            return fail(c, "/apoth affix must be executed by a living entity.", -10);
        }
        final LivingEntity living = (LivingEntity)entity;
        final ItemStack held = living.m_21205_();
        if (held.m_41619_()) {
            return fail(c, "The target entity must have an item in their main hand.", -2);
        }
        final DynamicHolder<LootRarity> rarity = (DynamicHolder<LootRarity>)AffixHelper.getRarity(held);
        if (!rarity.isBound()) {
            return fail(c, "The target item must have a set rarity.", -3);
        }
        final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(held);
        ((CommandSourceStack)c.getSource()).m_243053_((Component)Component.m_237110_("Affixes present on %s:", new Object[] { held.m_41611_() }));
        affixes.forEach((afx, inst) -> {
            final MutableComponent name = Component.m_237110_("[%s]", new Object[] { inst.getName(true) });
            name.m_6270_(Style.f_131099_.m_131140_(ChatFormatting.YELLOW).m_131144_(new HoverEvent(HoverEvent.Action.f_130831_, (Object)inst.getAugmentingText())));
            ((CommandSourceStack)c.getSource()).m_243053_((Component)Component.m_237110_("%s - %s%%", new Object[] { name, Affix.fmt(100.0f * inst.level()) }));
            return;
        });
        return 0;
    }
    
    public static int listAlternatives(final CommandContext<CommandSourceStack> c, final ResourceLocation affixId) {
        final DynamicHolder<Affix> afx = (DynamicHolder<Affix>)AffixRegistry.INSTANCE.holder(affixId);
        if (!afx.isBound()) {
            return fail(c, "Unknown affix: " + affixId, -1);
        }
        final Entity entity = ((CommandSourceStack)c.getSource()).m_81373_();
        if (!(entity instanceof LivingEntity)) {
            return fail(c, "/apoth affix must be executed by a living entity.", -10);
        }
        final LivingEntity living = (LivingEntity)entity;
        final ItemStack held = living.m_21205_();
        if (held.m_41619_()) {
            return fail(c, "The target entity must have an item in their main hand.", -2);
        }
        final DynamicHolder<LootRarity> rarity = (DynamicHolder<LootRarity>)AffixHelper.getRarity(held);
        if (!rarity.isBound()) {
            return fail(c, "The target item must have a set rarity.", -3);
        }
        final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(held);
        if (!affixes.containsKey(afx)) {
            return fail(c, "The target item does not contain the selected affix.", -4);
        }
        final List<DynamicHolder<? extends Affix>> alternatives = LootController.getAvailableAffixes(held, (LootRarity)rarity.get(), (Set)affixes.keySet(), ((Affix)afx.get()).getType());
        ((CommandSourceStack)c.getSource()).m_243053_((Component)Component.m_237110_("Possible alternatives to %s:", new Object[] { ((Affix)afx.get()).getName(true) }));
        alternatives.forEach(a -> {
            final MutableComponent name = Component.m_237110_("[%s]", new Object[] { ((Affix)a.get()).getName(true) });
            final Component augTxt = ((Affix)a.get()).getAugmentingText(held, (LootRarity)rarity.get(), 0.5f);
            name.m_6270_(Style.f_131099_.m_131140_(ChatFormatting.YELLOW).m_131144_(new HoverEvent(HoverEvent.Action.f_130831_, (Object)augTxt)));
            ((CommandSourceStack)c.getSource()).m_243053_((Component)AttributeHelper.list().m_7220_((Component)name));
            return;
        });
        return 0;
    }
    
    public static int fail(final CommandContext<CommandSourceStack> c, final String msg, final int code) {
        ((CommandSourceStack)c.getSource()).m_81352_((Component)Component.m_237115_(msg));
        return code;
    }
    
    static {
        SUGGEST_AFFIX = ((ctx, builder) -> SharedSuggestionProvider.m_82981_((Stream)AffixRegistry.INSTANCE.getKeys().stream().map(ResourceLocation::toString), builder));
        SUGGEST_APPLICABLE_AFFIX = ((ctx, builder) -> {
            final Entity entity = ((CommandSourceStack)ctx.getSource()).m_81373_();
            if (entity instanceof final LivingEntity living) {
                final ItemStack held = living.m_21205_();
                if (!held.m_41619_()) {
                    final LootCategory cat = LootCategory.forItem(held);
                    final DynamicHolder<LootRarity> rarity = (DynamicHolder<LootRarity>)AffixHelper.getRarity(held);
                    if (!cat.isNone() && rarity.isBound()) {
                        final Stream filter = AffixRegistry.INSTANCE.getValues().stream().filter(a -> a.canApplyTo(held, cat, (LootRarity)rarity.get()));
                        final AffixRegistry instance = AffixRegistry.INSTANCE;
                        Objects.requireNonNull(instance);
                        final Stream<String> suggestions = filter.map(instance::getKey).map(ResourceLocation::toString);
                        return SharedSuggestionProvider.m_82981_((Stream)suggestions, builder);
                    }
                }
            }
            return SharedSuggestionProvider.m_82970_((Iterable)Collections.emptyList(), builder);
        });
        SUGGEST_AFFIX_ON_ITEM = ((ctx, builder) -> {
            final Entity entity = ((CommandSourceStack)ctx.getSource()).m_81373_();
            if (entity instanceof final LivingEntity living) {
                final ItemStack held = living.m_21205_();
                if (!held.m_41619_()) {
                    final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(held);
                    return SharedSuggestionProvider.m_82981_((Stream)affixes.keySet().stream().map((Function<? super Object, ?>)DynamicHolder::getId).map((Function<? super Object, ?>)ResourceLocation::toString), builder);
                }
            }
            return SharedSuggestionProvider.m_82970_((Iterable)Collections.emptyList(), builder);
        });
    }
}
